;user's session in http look of alterator
(define-module (alterator http dialog)
  :use-module (ice-9 threads)
  :use-module (srfi srfi-2)
  :use-module (alterator algo)
  :use-module (alterator object)
  :use-module (alterator command)
  :use-module (alterator http session)
  :export (make-http-dialog))

(define-operation get-widget)

;process incoming loop requests
(define (dialog-process-request dlg cmd)
  (format #t "PROCESS-REQUEST:req=~A~%" cmd)
  (and (command? cmd)
       (let ((action (cond-cdr (command-arg-ref cmd 'action))))
         (and action
              (case (sure-symbol action)
                ((get) (get-command dlg))
                ((event) (event-command dlg cmd))
                (else (error "unknown action" action)))))))

;process "get url" request
(define (get-command dlg)
  (define-operation bubbling)
  
  (format #t "COMMAND: get: starting recreate bubbling~%")
  (bubbling dlg 'recreate))

;invisible set for "event" request
(define (do-silent-process state-description)
  (define-operation silent-set)

  (format #t "SILENT process for state=~A~%" state-description)
  (and-let* ((id (sure-symbol (car state-description)))
             (widget (get-widget (fluid-ref http-session) id))
             (attrs (cdr state-description)))
            (for-each (lambda (attr)
                        (let ((name (car attr))
                              (value (cdr attr)))
                          (format #t "... processing attribute ~A~%" name)
                          (apply silent-set widget name (string-splitting value #\;))))
                      attrs)))

;process "event emit" request
(define (event-command dlg cmd)
  (define-operation call)
  
  (format #t "COMMAND: event: starting event callback~%")
  (and-let* ((name (cond-cdr (command-arg-ref cmd 'name)))
             (id   (cond-cdr (command-arg-ref cmd 'widget-id)))
             (state (cond-cdr (command-arg-ref cmd 'state))))
            (format #t "...OK, all args ready~%")
            (for-each do-silent-process state);first process state here
	    (let ((widget (get-widget (fluid-ref http-session) id)))
	     (format #t "widget is ~A~%" widget)
	     (and widget (call widget (sure-symbol name)))))) ;then run appropriate callback


(define (make-http-dialog id parent-id)
  (define-operation reset-output)
  (define-operation start-thread)
  (define-operation signal-output)
  (define-operation read-input)
  (define-operation get-id)
  
  (format #t "creating dialog subobject~%")
  (let ((do-loop #t)
        (stop-mutex (make-mutex))
        (stop-condvar (make-condition-variable))
        (session (fluid-ref http-session))) ;also save our session: it's a constant and inaccessible from thread
    (object
     #f
     ((start self)
      (format #t "DIALOG:START~%")
      (start-thread self);start thread
      (format #t "DIALOG:AFTER-START-THREAD~%")
      (with-mutex stop-mutex (wait-condition-variable stop-condvar stop-mutex)));wair for stop
     ((start-thread self)
      (format #t "DIALOG:START-THREAD~%")
      (set! do-loop #t)
      (reset-output session);reset all previous messages in output queue
      (http-post-event 'action "clear");clear window content
      (begin-thread
       (with-fluids ((http-session session));restore fluid in thread
                    (format #t "starting dialog's main thread~%")
                    (let modal-loop ()
		      (format #t "THREAD: waiting for next message ... ~%")
                      (let* ((msg (read-input session)))
                        (format #t "THREAD: next message is ~A~%" msg)
                        (and msg (dialog-process-request self msg)))
                      (if do-loop
                          (begin (signal-output session)
                                 (modal-loop))
                          (with-mutex stop-mutex (signal-condition-variable stop-condvar))))
                    (and (string=? (sure-string parent-id) "") (begin
                                                   (signal-output session)
                                                   (close-http-session))))
	(format #t "THREAD: DEAD~%"))
      (format #t "DIALOG: notify output that we are ready....~%")
      (signal-output session)) ;notify server about available output events
      ((stop self)
       (format #t "DIALOG: stopping~%")
       (reset-output session);reset all previous messages in output queue
       (http-post-event 'action "clear")
       (set! do-loop #f)))))
